"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Model = exports.makeUserIdentity = void 0;
const events_1 = require("events");
const os = __importStar(require("os"));
const chalk_1 = require("chalk");
const node_opcua_client_1 = require("node-opcua-client");
const node_opcua_status_code_1 = require("node-opcua-status-code");
const node_opcua_pseudo_session_1 = require("node-opcua-pseudo-session");
const utils_1 = require("../utils/utils");
const extract_browse_path_1 = require("../utils/extract_browse_path");
const attributeKeys = [];
for (let i = 1; i <= node_opcua_client_1.AttributeIds.AccessLevelEx - 1; i++) {
    attributeKeys.push(node_opcua_client_1.AttributeIds[i]);
}
const data = {
    reconnectionCount: 0,
    tokenRenewalCount: 0,
    receivedBytes: 0,
    sentBytes: 0,
    sentChunks: 0,
    receivedChunks: 0,
    backoffCount: 0,
    transactionCount: 0,
};
function makeUserIdentity(argv) {
    let userIdentity = { type: node_opcua_client_1.UserTokenType.Anonymous }; // anonymous
    if (argv.userName && argv.password) {
        userIdentity = {
            type: node_opcua_client_1.UserTokenType.UserName,
            userName: argv.userName,
            password: argv.password,
        };
    }
    else if (argv.userCertificate && argv.userCertificatePrivateKey) {
        userIdentity = {
            type: node_opcua_client_1.UserTokenType.Certificate,
            certificateData: argv.userCertificate,
            privateKey: "todo",
        };
    }
    return userIdentity;
}
exports.makeUserIdentity = makeUserIdentity;
const hasComponentNodeId = (0, node_opcua_client_1.resolveNodeId)("HasComponent").toString();
const hasPropertyNodeId = (0, node_opcua_client_1.resolveNodeId)("HasProperty").toString();
const hasSubTypeNodeId = (0, node_opcua_client_1.resolveNodeId)("HasSubtype").toString();
const organizesNodeId = (0, node_opcua_client_1.resolveNodeId)("Organizes").toString();
function referenceToSymbol(ref) {
    // "+-->" // aggregate
    switch (ref.referenceTypeId.toString()) {
        case organizesNodeId:
            return "─o──";
        case hasComponentNodeId:
            return "──┼";
        case hasPropertyNodeId:
            return "──╫";
        case hasSubTypeNodeId:
            return "───▷";
        default:
            return "-->";
    }
}
function symbol(ref) {
    const s = " ";
    if (ref.typeDefinition.toString() === "ns=0;i=61") {
        return [(0, chalk_1.yellow)("[F]"), (0, chalk_1.yellow)("[F]")]; // ["🗀", "🗁"]; // "📁⧇Ⓞ"
    }
    switch (ref.nodeClass) {
        case node_opcua_client_1.NodeClass.Object:
            return [(0, chalk_1.cyanBright)("[O]"), (0, chalk_1.cyanBright)("[O]")];
        case node_opcua_client_1.NodeClass.Variable:
            return [(0, chalk_1.greenBright)("[V]"), (0, chalk_1.greenBright)("[V]")];
        case node_opcua_client_1.NodeClass.Method:
            return [(0, chalk_1.magenta)("[M]"), (0, chalk_1.magenta)("[M]")];
        case node_opcua_client_1.NodeClass.ObjectType:
            return [(0, chalk_1.bgCyanBright)("[O]"), (0, chalk_1.cyan)("[OT]")];
        case node_opcua_client_1.NodeClass.VariableType:
            return [(0, chalk_1.bgGreenBright)("[V]"), (0, chalk_1.yellow)("Ⓥ")];
        case node_opcua_client_1.NodeClass.ReferenceType:
            return [chalk_1.bgWhiteBright.black("[R]"), (0, chalk_1.yellowBright)("➾")];
        case node_opcua_client_1.NodeClass.DataType:
            return [(0, chalk_1.bgBlueBright)("[D]"), (0, chalk_1.bgBlueBright)("Ⓓ")];
        case node_opcua_client_1.NodeClass.View:
            return [(0, chalk_1.magentaBright)("[V]"), (0, chalk_1.magentaBright)("Ⓓ")];
    }
    return s;
}
class Model extends events_1.EventEmitter {
    constructor() {
        super();
        this.userIdentity = { type: node_opcua_client_1.UserTokenType.Anonymous };
        this.verbose = false;
        this.endpointUrl = "";
        this.monitoredItemsListData = [];
        this.clientAlarms = new node_opcua_client_1.ClientAlarmList();
        this.data = data;
    }
    initialize(endpoint, securityMode, securityPolicy, certificateFile, clientCertificateManager, applicationName, applicationUri) {
        return __awaiter(this, void 0, void 0, function* () {
            this.endpointUrl = this.endpointUrl;
            this.client = node_opcua_client_1.OPCUAClient.create({
                endpointMustExist: false,
                securityMode,
                securityPolicy,
                defaultSecureTokenLifetime: 40000,
                certificateFile,
                clientCertificateManager,
                applicationName,
                applicationUri,
                clientName: "Opcua-Commander-" + os.hostname(),
                keepSessionAlive: true,
            });
            this.client.on("send_request", function () {
                data.transactionCount++;
            });
            this.client.on("send_chunk", function (chunk) {
                data.sentBytes += chunk.length;
                data.sentChunks++;
            });
            this.client.on("receive_chunk", function (chunk) {
                data.receivedBytes += chunk.length;
                data.receivedChunks++;
            });
            this.client.on("backoff", function (number, delay) {
                data.backoffCount += 1;
                console.log((0, chalk_1.yellow)(`backoff  attempt #${number} retrying in ${delay / 1000.0} seconds`));
            });
            this.client.on("start_reconnection", () => {
                console.log((0, chalk_1.red)(" !!!!!!!!!!!!!!!!!!!!!!!!  Starting reconnection !!!!!!!!!!!!!!!!!!! " + this.endpointUrl));
            });
            this.client.on("connection_reestablished", () => {
                console.log((0, chalk_1.red)(" !!!!!!!!!!!!!!!!!!!!!!!!  CONNECTION RE-ESTABLISHED !!!!!!!!!!!!!!!!!!! " + this.endpointUrl));
                data.reconnectionCount++;
            });
            // monitoring des lifetimes
            this.client.on("lifetime_75", (token) => {
                if (this.verbose) {
                    console.log((0, chalk_1.red)("received lifetime_75 on " + this.endpointUrl));
                }
            });
            this.client.on("security_token_renewed", () => {
                data.tokenRenewalCount += 1;
                if (this.verbose) {
                    console.log((0, chalk_1.green)(" security_token_renewed on " + this.endpointUrl));
                }
            });
        });
    }
    create_subscription() {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.session) {
                throw new Error("Invalid Session");
            }
            const parameters = {
                requestedPublishingInterval: 500,
                requestedLifetimeCount: 1000,
                requestedMaxKeepAliveCount: 12,
                maxNotificationsPerPublish: 100,
                publishingEnabled: true,
                priority: 10,
            };
            try {
                this.subscription = yield this.session.createSubscription2(parameters);
                console.log("subscription created");
            }
            catch (err) {
                console.log("Cannot create subscription");
            }
        });
    }
    doConnect(endpointUrl, userIdentity) {
        return __awaiter(this, void 0, void 0, function* () {
            this.userIdentity = userIdentity;
            console.log("connecting to ....", endpointUrl);
            try {
                yield this.client.connect(endpointUrl);
            }
            catch (err) {
                console.log(" Cannot connect", err.toString());
                if (this.client.securityMode !== node_opcua_client_1.MessageSecurityMode.None && err.message.match(/has been disconnected by third party/)) {
                    console.log("Because you are using a secure connection, you need to make sure that the certificate\n" +
                        "of opcua-commander is trusted by the server you're trying to connect to.\n" +
                        "Please see the documentation for instructions on how to import a certificate into the CA store of the server.\n" +
                        `The opcua-commander certificate is in the folder \n${(0, chalk_1.cyan)(this.client.certificateFile)}`);
                }
                this.emit("connectionError", err);
                return;
            }
            try {
                this.session = yield this.client.createSession(this.userIdentity);
            }
            catch (err) {
                console.log(" Cannot create session ", err.toString());
                console.log((0, chalk_1.red)("  exiting"));
                setTimeout(function () {
                    return process.exit(-1);
                }, 25000);
                return;
            }
            this.session.on("session_closed", () => {
                console.log(" Warning => Session closed");
            });
            this.session.on("keepalive", () => {
                console.log("session keepalive");
            });
            this.session.on("keepalive_failure", () => {
                console.log("session keepalive failure");
            });
            console.log("connected to ....", endpointUrl);
            yield this.create_subscription();
        });
    }
    disconnect() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.session) {
                const session = this.session;
                this.session = undefined;
                yield session.close();
            }
            yield this.client.disconnect();
        });
    }
    request_write_item(treeItem) {
        if (!this.subscription)
            return;
        const node = treeItem.node;
        return treeItem;
    }
    writeNode(node, data) {
        return __awaiter(this, void 0, void 0, function* () {
            const dataTypeIdDataValue = yield this.session.read({ nodeId: node.nodeId, attributeId: node_opcua_client_1.AttributeIds.DataType });
            const arrayDimensionDataValue = yield this.session.read({ nodeId: node.nodeId, attributeId: node_opcua_client_1.AttributeIds.ArrayDimensions });
            const valueRankDataValue = yield this.session.read({ nodeId: node.nodeId, attributeId: node_opcua_client_1.AttributeIds.ValueRank });
            const dataTypeId = dataTypeIdDataValue.value.value;
            const dataType = yield (0, node_opcua_pseudo_session_1.findBasicDataType)(this.session, dataTypeId);
            const arrayDimension = arrayDimensionDataValue.value.value;
            const valueRank = valueRankDataValue.value.value;
            const coerceBoolean = (data) => {
                return data === "true" || data === "1" || data === true;
            };
            const coerceNumber = (data) => {
                return parseInt(data, 10);
            };
            const coerceNumberR = (data) => {
                return parseFloat(data);
            };
            const coerceNoop = (data) => data;
            const coerceFunc = (dataType) => {
                switch (dataType) {
                    case node_opcua_client_1.DataType.Boolean:
                        return coerceBoolean;
                    case node_opcua_client_1.DataType.Int16:
                    case node_opcua_client_1.DataType.Int32:
                    case node_opcua_client_1.DataType.Int64:
                    case node_opcua_client_1.DataType.UInt16:
                    case node_opcua_client_1.DataType.UInt32:
                    case node_opcua_client_1.DataType.UInt64:
                        return coerceNumber;
                    case node_opcua_client_1.DataType.Double:
                    case node_opcua_client_1.DataType.Float:
                        return coerceNumberR;
                    default:
                        return coerceNoop;
                }
            };
            if (dataType) {
                try {
                    const arrayType = valueRank === -1 ? node_opcua_client_1.VariantArrayType.Scalar : valueRank === 1 ? node_opcua_client_1.VariantArrayType.Array : node_opcua_client_1.VariantArrayType.Matrix;
                    const dimensions = arrayType === node_opcua_client_1.VariantArrayType.Matrix ? arrayDimension : undefined;
                    function coerceStringToDataType(data) {
                        const c = coerceFunc(dataType);
                        if (arrayType === node_opcua_client_1.VariantArrayType.Scalar) {
                            return c(data);
                        }
                        else {
                            return data.map((d) => c(d));
                        }
                    }
                    const value = new node_opcua_client_1.Variant({
                        dataType,
                        arrayType,
                        dimensions,
                        value: coerceStringToDataType(data),
                    });
                    const writeValue = new node_opcua_client_1.WriteValue({
                        nodeId: node.nodeId,
                        attributeId: node_opcua_client_1.AttributeIds.Value,
                        value: {
                            value,
                        },
                    });
                    let statusCode = yield this.session.write(writeValue);
                    console.log("writing    ", writeValue.toString());
                    console.log("statusCode ", statusCode.toString());
                    this.emit("nodeChanged", node.nodeId);
                    return statusCode;
                }
                catch (err) {
                    return node_opcua_status_code_1.StatusCodes.BadInternalError;
                }
            }
            return false;
        });
    }
    extractBrowsePath(nodeId) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield (0, extract_browse_path_1.extractBrowsePath)(this.session, nodeId);
        });
    }
    readNode(node) {
        return __awaiter(this, void 0, void 0, function* () {
            return yield this.session.read(node);
        });
    }
    readNodeValue(node) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.session) {
                return null;
            }
            const dataValues = yield this.readNode(node);
            if (dataValues.statusCode == node_opcua_status_code_1.StatusCodes.Good) {
                if (dataValues.value.value) {
                    switch (dataValues.value.arrayType) {
                        case node_opcua_client_1.VariantArrayType.Scalar:
                            return "" + dataValues.value.value;
                        case node_opcua_client_1.VariantArrayType.Array:
                            return dataValues.value.value.join(",");
                        default:
                            return "";
                    }
                }
            }
            return null;
        });
    }
    monitor_item(treeItem) {
        if (!this.subscription)
            return;
        const node = treeItem.node;
        this.subscription.monitor({
            nodeId: node.nodeId,
            attributeId: node_opcua_client_1.AttributeIds.Value,
            //, dataEncoding: { namespaceIndex: 0, name:null }
        }, {
            samplingInterval: 1000,
            discardOldest: true,
            queueSize: 100,
        }, node_opcua_client_1.TimestampsToReturn.Both, node_opcua_client_1.MonitoringMode.Reporting, (err, monitoredItem) => {
            if (err) {
                console.log("cannot create monitored item", err.message);
                return;
            }
            node.monitoredItem = monitoredItem;
            const monitoredItemData = [node.displayName, node.nodeId.toString(), "Q"];
            this.monitoredItemsListData.push(monitoredItemData);
            this.emit("monitoredItemListUpdated", this.monitoredItemsListData);
            //   xxx                monitoredItemsList.setRows(monitoredItemsListData);
            monitoredItem.on("changed", (dataValue) => {
                console.log(" value ", node.browseName, node.nodeId.toString(), " changed to ", (0, chalk_1.green)(dataValue.value.toString()));
                if (dataValue.value.value.toFixed) {
                    node.valueAsString = (0, utils_1.w)(dataValue.value.value.toFixed(3), 16, " ");
                }
                else {
                    node.valueAsString = (0, utils_1.w)(dataValue.value.value.toString(), 16, " ");
                }
                monitoredItemData[2] = node.valueAsString;
                this.emit("monitoredItemChanged", this.monitoredItemsListData, node, dataValue);
            });
        });
    }
    unmonitor_item(treeItem) {
        const node = treeItem.node;
        // terminate subscription
        node.monitoredItem.terminate(() => {
            let index = -1;
            this.monitoredItemsListData.forEach((entry, i) => {
                if (entry[1] == node.nodeId.toString()) {
                    index = i;
                }
            });
            if (index > -1) {
                this.monitoredItemsListData.splice(index, 1);
            }
            node.monitoredItem = null;
            this.emit("monitoredItemListUpdated", this.monitoredItemsListData);
        });
    }
    installAlarmMonitoring() {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.session) {
                return;
            }
            this.clientAlarms = yield (0, node_opcua_client_1.installAlarmMonitoring)(this.session);
            this.clientAlarms.on("alarmChanged", () => {
                this.clientAlarms.purgeUnusedAlarms();
                this.emit("alarmChanged", this.clientAlarms);
            });
        });
    }
    readNodeAttributes(nodeId) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.session) {
                return [];
            }
            const nodesToRead = attributeKeys.map((attributeId) => ({
                nodeId,
                attributeId: (node_opcua_client_1.AttributeIds[attributeId]),
            }));
            try {
                const dataValues = yield this.session.read(nodesToRead);
                const results = [];
                for (let i = 0; i < nodesToRead.length; i++) {
                    const nodeToRead = nodesToRead[i];
                    const dataValue = dataValues[i];
                    if (dataValue.statusCode !== node_opcua_status_code_1.StatusCodes.Good) {
                        continue;
                    }
                    const s = toString1(nodeToRead.attributeId, dataValue);
                    results.push({
                        attribute: attributeIdToString[nodeToRead.attributeId],
                        text: s,
                    });
                }
                return results;
            }
            catch (err) {
                console.log(err);
                return [];
            }
        });
    }
    expand_opcua_node(node) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.session) {
                throw new Error("No Session yet");
            }
            if (this.session.isReconnecting) {
                throw new Error("Session is not available (reconnecting)");
            }
            const children = [];
            const nodesToBrowse = [
                {
                    nodeId: node.nodeId,
                    referenceTypeId: "Organizes",
                    includeSubtypes: true,
                    browseDirection: node_opcua_client_1.BrowseDirection.Forward,
                    resultMask: 0x3f,
                },
                {
                    nodeId: node.nodeId,
                    referenceTypeId: "Aggregates",
                    includeSubtypes: true,
                    browseDirection: node_opcua_client_1.BrowseDirection.Forward,
                    resultMask: 0x3f,
                },
                {
                    nodeId: node.nodeId,
                    referenceTypeId: "HasSubtype",
                    includeSubtypes: true,
                    browseDirection: node_opcua_client_1.BrowseDirection.Forward,
                    resultMask: 0x3f,
                },
            ];
            try {
                const results = yield this.session.browse(nodesToBrowse);
                // organized
                let result = results[0];
                if (result.references) {
                    for (let i = 0; i < result.references.length; i++) {
                        const ref = result.references[i];
                        children.push({
                            arrow: referenceToSymbol(ref) + symbol(ref)[0],
                            displayName: ref.displayName.text || ref.browseName.toString(),
                            nodeId: ref.nodeId,
                            nodeClass: ref.nodeClass,
                        });
                    }
                }
                // Aggregates
                result = results[1];
                if (result.references) {
                    for (let i = 0; i < result.references.length; i++) {
                        const ref = result.references[i];
                        children.push({
                            arrow: referenceToSymbol(ref) + symbol(ref)[0],
                            displayName: ref.displayName.text || ref.browseName.toString(),
                            nodeId: ref.nodeId,
                            nodeClass: ref.nodeClass,
                        });
                    }
                }
                // HasSubType
                result = results[2];
                if (result.references) {
                    for (let i = 0; i < result.references.length; i++) {
                        const ref = result.references[i];
                        children.push({
                            arrow: referenceToSymbol(ref) + symbol(ref)[0],
                            displayName: ref.displayName.text || ref.browseName.toString(),
                            nodeId: ref.nodeId,
                            nodeClass: ref.nodeClass,
                        });
                    }
                }
                return children;
            }
            catch (err) {
                console.log(err);
                return [];
            }
        });
    }
}
exports.Model = Model;
function invert(o) {
    const r = {};
    for (const [k, v] of Object.entries(o)) {
        r[v.toString()] = k;
    }
    return r;
}
const attributeIdToString = invert(node_opcua_client_1.AttributeIds);
const DataTypeIdsToString = invert(node_opcua_client_1.DataTypeIds);
function dataValueToString(dataValue) {
    if (!dataValue.value || dataValue.value.value === null) {
        return "<???> : " + dataValue.statusCode.toString();
    }
    switch (dataValue.value.arrayType) {
        case node_opcua_client_1.VariantArrayType.Scalar:
            return dataValue.toString();
        case node_opcua_client_1.VariantArrayType.Array:
            return dataValue.toString();
        default:
            return "";
    }
}
function toString1(attribute, dataValue) {
    if (!dataValue || !dataValue.value || !dataValue.value.hasOwnProperty("value")) {
        return "<null>";
    }
    switch (attribute) {
        case node_opcua_client_1.AttributeIds.DataType:
            return DataTypeIdsToString[dataValue.value.value.value] + " (" + dataValue.value.value.toString() + ")";
        case node_opcua_client_1.AttributeIds.NodeClass:
            return node_opcua_client_1.NodeClass[dataValue.value.value] + " (" + dataValue.value.value + ")";
        case node_opcua_client_1.AttributeIds.IsAbstract:
        case node_opcua_client_1.AttributeIds.Historizing:
        case node_opcua_client_1.AttributeIds.EventNotifier:
            return dataValue.value.value ? "true" : "false";
        case node_opcua_client_1.AttributeIds.WriteMask:
        case node_opcua_client_1.AttributeIds.UserWriteMask:
            return " (" + dataValue.value.value + ")";
        case node_opcua_client_1.AttributeIds.NodeId:
        case node_opcua_client_1.AttributeIds.BrowseName:
        case node_opcua_client_1.AttributeIds.DisplayName:
        case node_opcua_client_1.AttributeIds.Description:
        case node_opcua_client_1.AttributeIds.ValueRank:
        case node_opcua_client_1.AttributeIds.ArrayDimensions:
        case node_opcua_client_1.AttributeIds.Executable:
        case node_opcua_client_1.AttributeIds.UserExecutable:
        case node_opcua_client_1.AttributeIds.MinimumSamplingInterval:
            if (!dataValue.value.value) {
                return "null";
            }
            return dataValue.value.value.toString();
        case node_opcua_client_1.AttributeIds.UserAccessLevel:
        case node_opcua_client_1.AttributeIds.AccessLevel:
            if (!dataValue.value.value) {
                return "null";
            }
            return (0, node_opcua_client_1.accessLevelFlagToString)(dataValue.value.value) + " (" + dataValue.value.value + ")";
        default:
            return dataValueToString(dataValue);
    }
}
//# sourceMappingURL=model.js.map