"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.extractBrowsePath = void 0;
const node_opcua_client_1 = require("node-opcua-client");
function readBrowseName(session, nodeId) {
    return __awaiter(this, void 0, void 0, function* () {
        const node = yield session.read({ nodeId, attributeId: node_opcua_client_1.AttributeIds.BrowseName });
        return node.value.value;
    });
}
function getParent(session, nodeId) {
    var _a, _b;
    return __awaiter(this, void 0, void 0, function* () {
        let browseResult = yield session.browse({
            browseDirection: node_opcua_client_1.BrowseDirection.Inverse,
            includeSubtypes: true,
            nodeId,
            nodeClassMask: 0xff,
            resultMask: 0xff,
            referenceTypeId: "HasChild",
        });
        if (browseResult.statusCode === node_opcua_client_1.StatusCodes.Good && ((_a = browseResult.references) === null || _a === void 0 ? void 0 : _a.length)) {
            const parentNodeId = browseResult.references[0].nodeId;
            return { sep: ".", parentNodeId };
        }
        browseResult = yield session.browse({
            browseDirection: node_opcua_client_1.BrowseDirection.Inverse,
            includeSubtypes: true,
            nodeId,
            nodeClassMask: 0xff,
            resultMask: 0xff,
            referenceTypeId: "Organizes",
        });
        if (browseResult.statusCode === node_opcua_client_1.StatusCodes.Good && ((_b = browseResult.references) === null || _b === void 0 ? void 0 : _b.length)) {
            const parentNodeId = browseResult.references[0].nodeId;
            return { sep: "/", parentNodeId };
        }
        return null;
    });
}
function extractBrowsePath(session, nodeId) {
    var _a, _b;
    return __awaiter(this, void 0, void 0, function* () {
        const browseName = yield readBrowseName(session, nodeId);
        const pathElements = [];
        pathElements.push(`${browseName.namespaceIndex}:${browseName.name}`);
        let parent = yield getParent(session, nodeId);
        while (parent) {
            if ((0, node_opcua_client_1.sameNodeId)(parent.parentNodeId, (0, node_opcua_client_1.resolveNodeId)("RootFolder"))) {
                break;
            }
            const browseName = yield readBrowseName(session, parent.parentNodeId);
            pathElements.unshift(`${browseName.namespaceIndex}:${browseName.name}${parent.sep}`);
            parent = yield getParent(session, parent.parentNodeId);
        }
        const browsePath = "/" + pathElements.join("");
        // verification
        const a = yield session.translateBrowsePath((0, node_opcua_client_1.makeBrowsePath)("i=84", browsePath));
        return browsePath + " (" + ((_b = (_a = a.targets[0]) === null || _a === void 0 ? void 0 : _a.targetId) === null || _b === void 0 ? void 0 : _b.toString()) + ")";
    });
}
exports.extractBrowsePath = extractBrowsePath;
//# sourceMappingURL=extract_browse_path.js.map