"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    var desc = Object.getOwnPropertyDescriptor(m, k);
    if (!desc || ("get" in desc ? !m.__esModule : desc.writable || desc.configurable)) {
      desc = { enumerable: true, get: function() { return m[k]; } };
    }
    Object.defineProperty(o, k2, desc);
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.View = exports.makeItems = void 0;
const blessed = __importStar(require("blessed"));
const util_1 = require("util");
const chalk_1 = __importDefault(require("chalk"));
const tree_item_1 = require("../widget/tree_item");
const node_opcua_client_1 = require("node-opcua-client");
const widget_tree_1 = require("../widget/widget_tree");
const alarm_box_1 = require("./alarm_box");
const utils_1 = require("../utils/utils");
const w2 = "40%";
const scrollbar = {
    ch: " ",
    track: {
        bg: "cyan",
    },
    style: {
        inverse: true,
    },
};
const style = {
    focus: {
        border: {
            fg: "yellow",
        },
        bold: false,
    },
    item: {
        hover: {
            bg: "blue",
        },
    },
    selected: {
        bg: "blue",
        bold: true,
    },
};
let old_console_log;
function makeItems(arr, width) {
    return arr.map((a) => {
        return (0, utils_1.w)(a[0], 25, ".") + ": " + (0, utils_1.w)(a[1], width, " ");
    });
}
exports.makeItems = makeItems;
let refreshTimer = null;
class View {
    constructor(model) {
        this.$headers = [];
        this.model = model;
        // Create a screen object.
        this.screen = blessed.screen({
            smartCSR: true,
            autoPadding: false,
            fullUnicode: true,
            title: "OPCUA CLI-Client",
        });
        // create the main area
        this.area1 = blessed.box({
            top: 0,
            left: 0,
            width: "100%",
            height: "90%-10",
        });
        this.area2 = blessed.box({
            top: "90%-9",
            left: 0,
            width: "100%",
            height: "shrink",
        });
        this.screen.append(this.area1);
        this.screen.append(this.area2);
        this.attributeList = this.install_attributeList();
        this.install_monitoredItemsWindow();
        this.install_writeFormWindow();
        this.logWindow = this.install_logWindow();
        this.menuBar = this.install_mainMenu();
        this.tree = this.install_address_space_explorer();
        // Render the screen.
        this.screen.render();
    }
    install_writeFormWindow() {
        this.writeForm = blessed.box({
            parent: this.area1,
            tags: true,
            top: "50%",
            left: w2 + "+1",
            width: "60%-1",
            height: "50%",
            keys: true,
            mouse: true,
            label: " Write item ",
            border: "line",
            scrollbar: scrollbar,
            noCellBorders: true,
            style: Object.assign({}, style),
            align: "left",
            hidden: true,
        });
        {
            const form = blessed.form({
                parent: this.writeForm,
                width: "100%-2",
                height: "100%-2",
                top: 1,
                left: 1,
                keys: true,
            });
            blessed.text({
                parent: form,
                top: 0,
                left: 0,
                content: "VALUES (Comma separated for array):",
            });
            this.valuesToWriteElement = blessed.textbox({
                parent: form,
                name: "valuesToWrite",
                top: 1,
                left: 0,
                height: "100%-2",
                inputOnFocus: true,
                mouse: false,
                vi: false,
                keys: false,
                content: "",
                border: {
                    type: "line",
                },
                focus: {
                    fg: "blue",
                },
            });
            const padding = {
                top: 0,
                right: 2,
                bottom: 0,
                left: 2,
            };
            const buttonTop = "100%-1";
            var submit = blessed.button({
                parent: form,
                name: "submit",
                content: "Submit",
                top: buttonTop,
                left: 0,
                shrink: true,
                mouse: true,
                padding,
                style: {
                    bold: true,
                    fg: "white",
                    bg: "green",
                    focus: {
                        inverse: true,
                    },
                },
            });
            submit.on("press", function () {
                form.submit();
            });
            var closeForm = blessed.button({
                parent: form,
                name: "close",
                content: "close",
                top: buttonTop,
                right: 0,
                shrink: true,
                mouse: true,
                padding,
                style: {
                    bold: true,
                    fg: "white",
                    bg: "red",
                    focus: {
                        inverse: true,
                    },
                },
            });
            closeForm.on("press", () => {
                this.writeForm.hide();
                this.screen.render();
            });
            const writeResultMsg = blessed.text({
                parent: form,
                top: submit.top,
                left: "center",
                content: "",
            });
            form.on("submit", (data) => __awaiter(this, void 0, void 0, function* () {
                const treeItem = this.tree.getSelectedItem();
                if (treeItem.node) {
                    // check if it is an array
                    const dataValues = yield this.model.readNode(treeItem.node);
                    let valuesToWrite = data.valuesToWrite;
                    if (dataValues && dataValues.value) {
                        if (dataValues.value.arrayType == node_opcua_client_1.VariantArrayType.Array) {
                            // since it is an array I will split by comma
                            valuesToWrite = valuesToWrite.split(",");
                        }
                    }
                    // send data to opc
                    const res = yield this.model.writeNode(treeItem.node, valuesToWrite);
                    if (res.valueOf() == 0) {
                        writeResultMsg.setContent("Write successful");
                    }
                    else {
                        writeResultMsg.setContent("Write error");
                    }
                    this.screen.render();
                }
            }));
        }
        this.area1.append(this.writeForm);
    }
    install_monitoredItemsWindow() {
        this.monitoredItemsList = blessed.listtable({
            parent: this.area1,
            tags: true,
            top: "50%",
            left: w2 + "+1",
            width: "60%-1",
            height: "50%",
            keys: true,
            label: " Monitored Items ",
            border: "line",
            scrollbar: scrollbar,
            noCellBorders: true,
            style: Object.assign({}, style),
            align: "left",
        });
        this.area1.append(this.monitoredItemsList);
        // binding .....
        this.model.on("monitoredItemListUpdated", (monitoredItemsListData) => {
            if (monitoredItemsListData.length > 0) {
                this.monitoredItemsList.setRows(monitoredItemsListData);
            }
            else {
                // when using setRows with empty array, the view does not update.
                // setting an empty row.
                const empty = [[" "]];
                this.monitoredItemsList.setRows(empty);
            }
            this.monitoredItemsList.render();
        });
        this.model.on("monitoredItemChanged", this._onMonitoredItemChanged.bind(this));
        this.model.on("nodeChanged", this._onNodeChanged.bind(this));
    }
    _onMonitoredItemChanged(monitoredItemsListData /*node: any, dataValue: DataValue*/) {
        this.monitoredItemsList.setRows(monitoredItemsListData);
        this.monitoredItemsList.render();
    }
    install_logWindow() {
        const logWindow = blessed.list({
            parent: this.area2,
            tags: true,
            label: " {bold}{cyan-fg}Info{/cyan-fg}{/bold} ",
            top: "top",
            left: "left",
            width: "100%",
            height: "100%-2",
            keys: true,
            border: "line",
            scrollable: true,
            scrollbar: {
                ch: " ",
                track: {
                    bg: "cyan",
                },
                style: {
                    inverse: true,
                },
            },
            style: Object.assign({}, style),
        });
        old_console_log = console.log;
        console.log = function (...args) {
            const str = util_1.format.apply(null, args);
            const lines = str.split("\n");
            lines.forEach((str) => {
                logWindow.addItem(str);
            });
            logWindow.select(logWindow.items.length - 1);
        };
        this.area2.append(logWindow);
        return logWindow;
    }
    install_mainMenu() {
        const menuBarOptions = {
            parent: this.area2,
            top: "100%-2",
            left: "left",
            width: "100%",
            height: 2,
            keys: true,
            style: Object.assign(Object.assign({}, style), { prefix: {
                    fg: "cyan",
                } }),
            //xx label: " {bold}{cyan-fg}Info{/cyan-fg}{/bold}",
            //xx border: "line",
            bg: "cyan",
            commands: [],
            items: [],
            autoCommandKeys: true,
        };
        const menuBar = blessed.listbar(menuBarOptions);
        this.area2.append(menuBar);
        menuBar.setItems({
            Monitor: {
                //xx prefix: "M",
                keys: ["m"],
                callback: () => this._onMonitoredSelectedItem(),
            },
            Write: {
                keys: ["w"],
                callback: () => this._onWriteSelectedItem(),
            },
            Exit: {
                keys: ["q", "x"],
                callback: () => this._onExit(),
            },
            Tree: {
                keys: ["t"],
                callback: () => this.tree.focus(),
            },
            Attributes: {
                keys: ["l"],
                callback: () => this.attributeList.focus(),
            },
            Info: {
                keys: ["i"],
                callback: () => this.logWindow.focus(),
            },
            Clear: {
                keys: ["c"],
                callback: () => {
                    this.logWindow.clearItems();
                    this.logWindow.screen.render();
                },
            },
            Unmonitor: {
                keys: ["u"],
                callback: () => this._onUnmonitoredSelectedItem(),
            },
            Stat: {
                keys: ["s"],
                callback: () => this._onDumpStatistics(),
            },
            Alarm: {
                keys: ["a"],
                callback: this._onToggleAlarmWindows.bind(this),
            },
            //  "Menu": { keys: ["A-a", "x"], callback: () => this.menuBar.focus() }
        });
        return menuBar;
    }
    install_address_space_explorer() {
        this.tree = new widget_tree_1.Tree({
            parent: this.area1,
            tags: true,
            fg: "green",
            //Xx keys: true,
            label: " {bold}{cyan-fg}Address Space{/cyan-fg}{/bold} ",
            top: "top",
            left: "left",
            width: "40%",
            height: "100%",
            keys: true,
            vi: true,
            mouse: true,
            border: "line",
            style: Object.assign({}, style),
        });
        //allow control the table with the keyboard
        this.tree.on("select", (treeItem) => {
            if (treeItem) {
                this.fill_attributesRegion(treeItem.node.nodeId);
            }
        });
        this.tree.on("keypress", (ch, key) => {
            if (key.name === "up" || key.name === "down") {
                if (refreshTimer) {
                    return;
                }
                refreshTimer = setTimeout(() => {
                    const treeItem = this.tree.getSelectedItem();
                    if (treeItem && treeItem.node) {
                        this.fill_attributesRegion(treeItem.node.nodeId);
                    }
                    refreshTimer = null;
                }, 100);
            }
        });
        this.area1.append(this.tree);
        this.populateTree();
        this.tree.focus();
        return this.tree;
    }
    populateTree() {
        this.tree.setData({
            name: "RootFolder",
            nodeId: (0, node_opcua_client_1.resolveNodeId)("RootFolder"),
            children: this.expand_opcua_node.bind(this),
        });
    }
    expand_opcua_node(node, callback) {
        function f(node) {
            return __awaiter(this, void 0, void 0, function* () {
                try {
                    let children = yield this.model.expand_opcua_node(node);
                    // we sort the childrens by displayName alphabetically
                    children = children.sort((a, b) => {
                        return a.displayName < b.displayName ? -1 : a.displayName > b.displayName ? 1 : 0;
                    });
                    const results = children.map((c) => new tree_item_1.TreeItem(Object.assign(Object.assign({}, c), { children: this.expand_opcua_node.bind(this) })));
                    return results;
                }
                catch (err) {
                    throw new Error("cannot expand");
                }
            });
        }
        (0, util_1.callbackify)(f).call(this, node, callback);
    }
    _onNodeChanged(nodeId) {
        if ((0, node_opcua_client_1.sameNodeId)(this.attributeListNodeId, nodeId)) {
            // we need to refresh the attribute list
            this.fill_attributesRegion(nodeId);
        }
    }
    fill_attributesRegion(nodeId) {
        return __awaiter(this, void 0, void 0, function* () {
            const attr = [];
            function append_text(prefix, s, attr) {
                const a = s.split("\n");
                if (a.length === 1) {
                    attr.push([prefix, s]);
                }
                else {
                    attr.push([prefix, a[0]]);
                    for (let j = 1; j < a.length; j++) {
                        attr.push(["   |    ", a[j]]);
                    }
                }
            }
            const attributes = yield this.model.readNodeAttributes(nodeId);
            if (attributes.length === 0) {
                return;
            }
            for (const r of attributes) {
                append_text(r.attribute, r.text, attr);
            }
            const width = this.attributeList.width - 28;
            this.attributeList.setItems(makeItems(attr, width));
            this.attributeList.screen.render();
            this.attributeListNodeId = nodeId;
        });
    }
    install_attributeList() {
        this.attributeList = blessed.list({
            parent: this.area1,
            label: " {bold}{cyan-fg}Attribute List{/cyan-fg}{/bold} ",
            top: 0,
            tags: true,
            left: w2 + "+1",
            width: "60%-1",
            height: "50%",
            border: "line",
            // noCellBorders: true,
            scrollbar: scrollbar,
            style: Object.assign({}, style),
            align: "left",
            keys: true,
        });
        this.area1.append(this.attributeList);
        const width = this.attributeList.width - 28;
        this.attributeList.setItems(makeItems([], width));
        return this.attributeList;
    }
    install_alarm_windows() {
        if (this.alarmBox) {
            this.alarmBox.show();
            this.alarmBox.focus();
            return;
        }
        this.alarmBox = blessed.listtable({
            parent: this.area1,
            tags: true,
            fg: "green",
            // label: "{bold}{cyan-fg}Alarms - Conditions {/cyan-fg}{/bold} ",
            label: "Alarms - Conditions",
            top: "top+6",
            left: "left+2",
            width: "100%-10",
            height: "100%-10",
            keys: true,
            border: "line",
            scrollbar: scrollbar,
            noCellBorders: false,
            style: Object.assign({}, style),
            align: "left"
        });
        this.$headers = [
            "EventType",
            "ConditionId",
            "SourceName",
            // "BranchId",
            // "EventId",
            "Message",
            "Severity",
            //"Enabled?", "Active?",  "Acked?", "Confirmed?", "Retain",
            "E!AC",
            "Comment",
        ];
        const data = [this.$headers];
        this.alarmBox.setData(data);
        this.model.installAlarmMonitoring();
        this.model.on("alarmChanged", (list) => (0, alarm_box_1.updateAlarmBox)(list, this.alarmBox, this.$headers));
        this.alarmBox.focus();
    }
    hide_alarm_windows() {
        this.alarmBox.hide();
    }
    _onExit() {
        return __awaiter(this, void 0, void 0, function* () {
            console.log(chalk_1.default.red(" disconnecting .... "));
            yield this.model.disconnect();
            console.log(chalk_1.default.green(" disconnected .... "));
            yield new Promise((resolve) => setTimeout(resolve, 1000));
            process.exit(0);
        });
    }
    _onToggleAlarmWindows() {
        return __awaiter(this, void 0, void 0, function* () {
            if (this.alarmBox && this.alarmBox.visible) {
                this.hide_alarm_windows();
            }
            else {
                this.install_alarm_windows();
                this.alarmBox.show();
            }
            this.screen.render();
        });
    }
    _onMonitoredSelectedItem() {
        const treeItem = this.tree.getSelectedItem();
        if (treeItem.node.monitoredItem) {
            console.log(" Already monitoring ", treeItem.node.nodeId.toString());
            return;
        }
        this.model.monitor_item(treeItem);
    }
    _onWriteSelectedItem() {
        return __awaiter(this, void 0, void 0, function* () {
            this.writeForm.show();
            const treeItem = this.tree.getSelectedItem();
            if (treeItem.node) {
                const treeItemToUse = this.model.request_write_item(treeItem);
                if (treeItemToUse) {
                    const value = yield this.model.readNodeValue(treeItem.node);
                    if (value) {
                        this.valuesToWriteElement.setValue(value);
                    }
                    else {
                        this.valuesToWriteElement.setValue("");
                    }
                    this.screen.render();
                    this.valuesToWriteElement.focus();
                    this.screen.render();
                }
                return;
            }
        });
    }
    _onUnmonitoredSelectedItem() {
        const treeItem = this.tree.getSelectedItem();
        if (!treeItem.node.monitoredItem) {
            console.log(treeItem.node.nodeId.toString(), " was not being monitored");
            return;
        }
        this.model.unmonitor_item(treeItem);
    }
    _onDumpStatistics() {
        return __awaiter(this, void 0, void 0, function* () {
            console.log("-----------------------------------------------------------------------------------------");
            console.log(chalk_1.default.green("     transaction count   : ", chalk_1.default.yellow(this.model.data.transactionCount)));
            console.log(chalk_1.default.green("            sent bytes   : ", chalk_1.default.yellow(this.model.data.sentBytes)));
            console.log(chalk_1.default.green("        received bytes   : ", chalk_1.default.yellow(this.model.data.receivedBytes)));
            console.log(chalk_1.default.green("   token renewal count   : ", chalk_1.default.yellow(this.model.data.tokenRenewalCount)));
            console.log(chalk_1.default.green("    reconnection count   : ", chalk_1.default.yellow(this.model.data.reconnectionCount)));
            console.log("-----------------------------------------------------------------------------------------");
            const treeItem = this.tree.getSelectedItem();
            const browsePath = yield this.model.extractBrowsePath(treeItem.node.nodeId);
            console.log(chalk_1.default.cyan("selected node browse path :", chalk_1.default.magenta(browsePath)));
        });
    }
}
exports.View = View;
//# sourceMappingURL=view.js.map