"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.Tree = void 0;
const chalk_1 = __importDefault(require("chalk"));
const node_opcua_client_1 = require("node-opcua-client");
const blessed = require("blessed");
// some unicode icon characters ►▼◊◌○●□▪▫֎☺◘♦
function isFunction(variableToCheck) {
    return variableToCheck instanceof Function;
}
function toContent(node, isLastChild, parent) {
    if (parent) {
        const sep = (parent.isLastChild) ? " " : "│";
        node.prefix = parent.prefix + sep;
    }
    else {
        node.prefix = " ";
    }
    const s = (isLastChild) ? "└" : "├";
    const level = node.depth;
    (0, node_opcua_client_1.assert)(level >= 0 && level < 100);
    const hasChildren = node.children && node.children.length > 0;
    //    [+]
    const c = node.expanded ? (hasChildren ? chalk_1.default.green("▼") : "─") : "►";
    const str = node.prefix + s + c + node.name;
    return str;
}
function dummy(node, callback) {
    callback(null, node.children);
}
class Tree extends blessed.List {
    constructor(options) {
        const scrollbar = {
            ch: " ",
            track: {
                bg: "cyan"
            },
            style: {
                inverse: true
            }
        };
        const style = {
            item: {
                hover: {
                    bg: "blue"
                }
            },
            selected: {
                bg: "blue",
                bold: true
            }
        };
        options.border = options.border || "line";
        options.scrollbar = options.scrollbar || scrollbar;
        options.style = options.style || style;
        options.keys = true;
        super(options);
        this.items = [];
        this.key(["+", "right"], this.expandSelected.bind(this));
        this.key(["-", "left"], this.collapseSelected.bind(this));
        this._index_selectedNode = 0;
    }
    _add(node, isLastChild, parent) {
        node.isLastChild = isLastChild;
        const item = this.add(toContent(node, isLastChild, parent));
        item.node = node;
        if (this._old_selectedNode === node) {
            this._index_selectedNode = this.itemCount - 1;
        }
    }
    get itemCount() { return this.items.length; }
    walk(node, depth) {
        if (this.itemCount) {
            this._old_selectedNode = this.getSelectedItem().node;
            (0, node_opcua_client_1.assert)(this._old_selectedNode);
        }
        this._index_selectedNode = -1;
        this.setItems([]);
        if (node.name && depth === 0) {
            // root node
            node.depth = 0;
            this._add(node, true, null);
        }
        function dumpChildren(node, depth) {
            if (isFunction(node.children)) {
                return;
            }
            node.children = node.children || [];
            let isLastChild;
            for (let i = 0; i < node.children.length; i++) {
                const child = node.children[i];
                if (child) {
                    child.depth = depth + 1;
                    isLastChild = (i === node.children.length - 1);
                    this._add(child, isLastChild, node);
                    if (child.expanded && !isFunction(child.children)) {
                        dumpChildren.call(this, child, depth + 1);
                    }
                }
            }
        }
        if (node.expanded) {
            dumpChildren.call(this, node, depth);
        }
        this._index_selectedNode = this._index_selectedNode >= 0 ? this._index_selectedNode : 0;
        this.select(this._index_selectedNode);
    }
    expandSelected() {
        const node = this.getSelectedItem().node;
        if (node.expanded) {
            return;
        }
        const populate_children = isFunction(node.children) ? node.children : dummy;
        populate_children.call(this, node, (err, children) => {
            if (err) {
                return;
            }
            (0, node_opcua_client_1.assert)(Array.isArray(children));
            node.children = children;
            node.expanded = true;
            this.setData(this.__data);
        });
    }
    collapseSelected() {
        const node = this.getSelectedItem().node;
        if (!node.expanded) {
            return;
        }
        node.expanded = false;
        this.setData(this.__data);
    }
    setData(data) {
        this.__data = data;
        this.walk(data, 0);
        this.screen.render();
    }
    getSelectedItem() {
        return this.getTreeItemAtPos(this.getSelectedIndex());
    }
    getTreeItemAtPos(selectedIndex) {
        return this.items[selectedIndex];
    }
    getSelectedIndex() {
        return this.selected;
    }
}
exports.Tree = Tree;
//# sourceMappingURL=widget_tree.js.map